c     Huifen Chen and Bruce Schmeiser, March 8, 2015.
c     Driver Program: MNO-PQRS
c       Max Nonnegativity Option---Piecewise-Quadratic Rate Smoothing
c     Purpose:  Fit the nonnegative rate function   
c       max{0, a_i x^2 + b_i x + c+i} to each interval, i=1,2,...,k.  
c       Maintain the k means rates_i, end-point continuities,  
c       end-point first derivatives, and maybe end rates.
      implicit doubleprecision (a-h,o-z)
      implicit integer (i-n)
      parameter (maxint = 10000)
      double precision, dimension (1:maxint) :: rates, a, b, c
      double precision, dimension (1:maxint) :: aplus, bplus, cplus
c     ..get number of intervals
      print *,' enter the number of intervals (1 to', maxint, ').'
      read (5,*) k
c     ..get rate for each interval
      print *,' enter the', k, ' nonnegative constant rates.'
      read (5,*) (rates(i),i=1,k)
c     ..get context information
      print *,' enter the context: 0,1,2,3, or 4.'
      print *,'   0: cyclic.'
      print *,'   1: finite horizon, two end-point rates specified.'
      print *,'   2: finite horizon, left end-point rate specified'
      print *,'   3: finite horizon, right end-point rate specified.'
      print *,'   4: finite horizon, no end-point rate specified.'
      read (5,*) icontxt
      rate0   = -1
      ratekp1 = -1
      if (icontxt .eq. 1) then
        print *,' enter the left and right end-point rates.'
        read (5,*) rate0, ratekp1
      elseif (icontxt .eq. 2) then
        print *,' enter the left end-point rate.'
        read (5,*) rate0
      elseif (icontxt .eq. 3) then
        print *,' enter the right end-point rate.'
        read (5,*) ratekp1
      endif        
c     ..echo input
      print *,' '
      print *,' Problem parameters...'
      print *,'   interval i, rates(i)'
      do i=1,k
        print *, i, rates(i)
      enddo
      print *,'   context   =', icontxt
      print *,'   end-point rates =', rate0, ratekp1
c     ..compute and report the unconstrained fit
      call PQRSfit( icontxt, k, rates, rate0, ratekp1, a, b, c )
      print *,' '
      print *,' PQRS fit (piecewise quadratic)...'
      print *,'   i, a(i), b(i), c(i), qrate'
      do i=1,k
        qrate = (a(i)/3) + (b(i)/2) + c(i)
        print *,  i, a(i), b(i), c(i), qrate
      enddo
c     ..compute and report the MNO "max" nonnegative fit
      call MNOfit( icontxt, k, rates, rate0, ratekp1, 
     &                            a, b, c, aplus, bplus, cplus )
      print *,' '
      print *,' MNO fit (piecewise-quadratic nonnegative)...'
        print *,'   i, aplus(i), bplus(i), cplus(i), integral'
      do i=1,k
        call qmaxint( aplus(i), bplus(i), cplus(i), qrate )
        print *, i, aplus(i), bplus(i), cplus(i), qrate
      enddo
      stop 999
      end


      subroutine PQRSfit( icontxt, k, rates, rate0, ratekp1, a, b, c )
c     Huifen Chen and Bruce Schmeiser, March 14, 2015.
c     Call the appropriate unconstrained fitting routine.
      implicit doubleprecision (a-h,o-z)
      implicit integer (i-n)
      double precision, dimension (1:k) :: rates, a, b, c
c     input: 
c       icontxt = horizon type in {0,1,2,3,4}
c       k       = number of intervals
c       rates   = specified rates, i=1,...,k
c       rate0   = left end-point rate
c       ratekp1 = right end-point rate
c     output:
c       (a,b,c) = piecewise-quadratic coefficients
      if (icontxt .eq. 0) then
c       ..context 0: cyclic
        call pqfit0( k, rates, a, b )
      elseif (icontxt .eq. 1) then
c       ..context 1: finite horizon, two end point rates
        call pqfit1( k, rates, rate0, ratekp1, a, b )
      elseif (icontxt .eq. 2) then
c       ..context 2: finite horizon, only left end-point rate
        call pqfit2( k, rates, rate0, a, b )
      elseif (icontxt .eq. 3) then
c       ..context 3: finite horizon, only right end-point rate
        call pqfit3( k, rates, ratekp1, a, b )
      elseif (icontxt .eq. 4) then
c       ..context 4: finite horizon, no end-point rate
        call pqfit4( k, rates, a, b )
      endif
c     ..given (a, b), compute c
      do i=1,k
        c(i) = rates(i) - ( a(i)/3 ) - ( b(i)/2 )
      enddo
      return
      end


      subroutine Mcell( k, irow, jcolumn, cell )
c     Huifen Chen and Bruce Schmeiser, March 14, 2015
c     Compute (closed-form) the (irow, icolumn) cell of
c       the cyclic-horizon (3k x 3k) M0 inverse. 
c     input: 
c       k = number of intervals, (k>1)
c       (irow, jcolumn) = matrix-inverse indices
c     output:
c       cell = M0^{-1} (irow, jcolumn) 
      implicit doubleprecision (a-h,o-z)
      implicit integer (i-n)
c     ..get equivalent cell indices (iindex, jindex) from rows 1,2,3
      iindex = irow - 3*((irow-1)/3)
      jindex = jcolumn - (irow - iindex)
      if (jindex .lt. 1) jindex = jindex + 3*k
c     ..get block number m and column jindex (=1,2,3) within the block
      m = (jindex - 1) / 3
      jindex = jindex - 3*m
c     ..compute the value of M0 inverse's (irow, icolumn) cell
      twopidk = 6.2831853071796d0 / k
      cell  = 0
      khalf = k / 2
      keven = 0
      if (k .eq. 2*khalf) keven = 1
      do j=0,khalf
        rj     = j * twopidk
        cosj   = cos( rj )
        factor = -1.5d0 / (2 + cosj)
        if (iindex .eq. 1) then
          cosjm   = cos( m*rj )
          if (jindex .eq. 1) rAj =  4 * (1 - cosj)       * cosjm
          if (jindex .eq. 2) rAj =  2 * (cos( (m+1)*rj ) - cosjm)
          if (jindex .eq. 3) rAj = -    (cos( (m+1)*rj ) + cosjm)
        elseif (iindex .eq. 2) then
          cosjmp1 = cos( (m+1)*rj )
          if (jindex .eq. 1) rAj =  4 *   (cosjmp1 - cos( m*rj ))
          if (jindex .eq. 2) rAj = -4 *    cosjmp1
          if (jindex .eq. 3) rAj =  2 * (2*cosjmp1 + cos( m*rj )) / 3
       else
          cosjmp1 = cos( (m+1)*rj )
          if (jindex .eq. 1) rAj = -2 * (  cosjmp1 + cos( m*rj ))
          if (jindex .eq. 2) rAj =  2 * (2*cosjmp1 + cos( m*rj )) / 3
          if (jindex .eq. 3) rAj = -       cosjmp1                / 3
        endif
        if (j .eq. 0 .or. (j.eq.khalf .and. keven.eq.1)) rAj = rAj / 2
        cell = cell + factor * rAj
      enddo
      cell = cell / k
      return
      end


      subroutine Scell( n, i, j, cell )
c     Huifen Chen and Bruce Schmeiser, March 14, 2015
c     Purpose: compute the (i,j) cell of S^{-1}, where S is 
c          the (3n x 3n) partitioned matrix with n blocks:
c               S = [S1 S2  0 0 ... 0   0]
c                   [0  S1 S2 0 ... 0   0]
c                    .                  .
c                   [0 ...          0   S1] ,
c          where each block matrix is 3x3 with
c                 S1 = [1/3 1/2 1; 1 1  1; 2  1 0], 
c                 S2 = [0   0   0; 0 0 -1; 0 -1 0], 
c          and the other entries are 0.  
c     input:
c       n = number of blocks
c       (i,j) = indices from S inverse
c     output: 
c       cell = the (i,j) cell of S inverse
      implicit doubleprecision (a-h,o-z)
      implicit integer (i-n)
      n3 = 3 * n
c     ..compute denominator
      call Mcell( n, 3 ,n3-1, cell0 )
      denom1 = 1 + cell0
c     ..compute cell (i,j) of E inverse
      call Mcell( n, i, j,    cell1 )
      call Mcell( n, i, n3-1, cell2 )
      call Mcell( n, 3, j,    cell3j )
      rEinvij = cell1 - (cell2 * cell3j / denom1)
c     ..compute cell (i,n3) of E inverse
      call Mcell( n, i, n3,   cell1 )
      call Mcell( n, 3, n3,   cell3n3 )
      rEinvin3 = cell1 - (cell2 * cell3n3 / denom1)
c     ..compute cell (2,j) of E inverse
      call Mcell( n, 2, j,    cell1 )
      call Mcell( n, 2, n3-1, cell2 )
      rEinv2j = cell1 - (cell2 * cell3j / denom1)
c     ..compute cell (2,n3) of E inverse
      call Mcell( n, 2, n3,   cell1 )
      rEinv2n3 = cell1 - (cell2 * cell3n3 / denom1)
c     ..compute the (i,j) cell of S inverse
      denom2 = 1 + rEinv2n3
      cell = rEinvij - (rEinvin3 * rEinv2j / denom2)
      return
      end


      subroutine pqfit0( k, rates, a, b )
c     Huifen Chen and Bruce Schmeiser, March 14, 2105.
c     PQRS Context: 0. Cyclic.
c     Fit unconstrained rates; i.e., ignore negativity.
c     Method:  Maintain k means, k continuity points, and
c              k first derivatives. Rows 1 and 2 of M0 inverse.
c     input: 
c       k     = number of intervals
c       rates = specified rates, i=1,...,k
c     output:
c       (a,b) = second- and first-order quadratic coefficients
      implicit doubleprecision (a-h,o-z)
      implicit integer (i-n)
      double precision, dimension (1:k) :: rates, a, b
c     ..initialize
      do i = 1,k
        a(i) = 0
        b(i) = 0 
      enddo
c     ..compute (a,b), block 0 to block k-1
      do m = 0,k-1
        jindex = 3*m + 1
        call Mcell( k, 1, jindex, cell1j )
        call Mcell( k, 2, jindex, cell2j )
        do i = 1,k
          jrotate = m + i
          if (jrotate .gt. k) jrotate = jrotate - k
          a(i) = a(i) + cell1j * rates(jrotate)
          b(i) = b(i) + cell2j * rates(jrotate)
        enddo
      enddo
      return
      end


      subroutine pqfit1( k, rates, rate0, ratekp1, a, b )
c     Huifen Chen and Bruce Schmeiser, March 14, 2015.
c     PQRS Context: 1. Finite horizon, two end-point rates.
c     Fit unconstrained rates; i.e., ignore negativity.
c     Method:  Maintain k means, k-1 continuity, k-1 first
c              derivatives, and two end-point rates. 
c     input: 
c       k     = number of intervals
c       rates = specified rates, i=1,...,k
c       rate0 = left end-point rate
c       ratekp1 = right end-point rate
c     output:
c       (a,b) = second- and first-order quadratic coefficients
      implicit doubleprecision (a-h,o-z)
      implicit integer (i-n)
      double precision, dimension (1:k) :: rates, a, b
      if (k .eq. 1) then
        a(1) = 3*(rate0 + ratekp1) - 6*rates(1)
        b(1) = ratekp1 - rate0 - a(1)
        return
      endif
c     ..initialize
      k3   = 3*k 
c     ..get constants p, q, and d1
      call Scell( k-1, 3, k3-3, p )
      call Scell( k-1, 3, k3-4, q )
      d1 = 4*p - q
c     ..compute (ai,bi), i=1,..,k-1
      do i=1,k-1
        a(i) = 0
        b(i) = 0
        do ioffset=0,1
          iindex = 3*i - 2 + ioffset
          call Scell( k-1, iindex, k3-3, rMuli3 )
          call Scell( k-1, iindex, k3-4, rMuli4 )
          do j=1,k-1
            jindex = 3*j-2
c           ..G11, the upper-left corner of M1inv
            call Scell( k-1, iindex, jindex, rMulij )
            call Scell( k-1, 3,      jindex, rMul3j )
            rMulij = rMulij - ((4*rMuli3 - rMuli4)*rMul3j) / d1
            if (ioffset .eq. 0) a(i) = a(i) + rMulij * rates(j)
            if (ioffset .eq. 1) b(i) = b(i) + rMulij * rates(j)
          enddo
c           ..G12, the upper-right corner of M1inv
          Call Scell( k-1, iindex, k3-3, celli3 )
          call Scell( k-1, iindex, k3-4, celli4 )
          rMuri0 =    (4*celli3 -   celli4) / d1
          rMuri1 =  2*(q*celli3 - p*celli4) / d1
          rMurik = -6*(q*celli3 - p*celli4) / d1
          if (ioffset .eq. 0) then
            a(i) = a(i) + rates(k) * rMurik
            a(i) = a(i) + rate0    * rMuri0
            a(i) = a(i) + ratekp1  * rMuri1
          else
            b(i) = b(i) + rates(k) * rMurik
            b(i) = b(i) + rate0    * rMuri0
            b(i) = b(i) + ratekp1  * rMuri1
          endif
        enddo
      enddo
c     ..compute (ak, bk)
      ckm1 = rates(k-1) - a(k-1)/3 - b(k-1)/2
      ck   = a(k-1) + b(k-1) + ckm1
      b(k) = 2*a(k-1) + b(k-1)
      a(k) = 3*(rates(k) - b(k)/2 - ck)
      return
      end


      subroutine pqfit2( k, rates, rate0, a, b )
c     Huifen Chen and Bruce Schmeiser, March 14, 2015.
c     PQRS Context: 2. Finite horizon, only left end-point rate.
c     Fit unconstrained rates; i.e., ignore negativity.
c     Method:  Minimize sum_i=1^k a_i^2, subject to k means, 
c              k continuity, and k-1 first derivatives.
c              a(k) is the free variable.
c     input: 
c       k     = number of intervals
c       rates = specified rates, i=1,...,k
c       rate0 = left end-point rate
c     output:
c       (a,b) = second- and first-order quadratic coefficients
      implicit doubleprecision (a-h,o-z)
      implicit integer (i-n)
      double precision, dimension (1:k) :: rates, a, b
      if (k .eq. 1) then
        a(1) = 0
        b(1) = 2*(rates(1) - rate0)
        return
      endif
c     ..initialize
      k3 = 3*k 
      bottom = 0
      top    = 0
c     ..use Q22, the lower-right (2x2) corner of sM2inv
      call Scell( k-1, 3, k3-3, p )
      call Scell( k-1, 3, k3-4, q )
      d2 = (q/2) - p
c     ..begin computing bi, i=1,..,k-1, using rows 2,5,... of Q11
      do i=1,k-1
        iindex = 3*i - 1
        b(i) = 0
        call Scell( k-1, iindex, k3-3, celli3 ) 
        call Scell( k-1, iindex, k3-4, celli4 ) 
        do j=1,k-1
          jindex= 3*j - 2
          call Scell( k-1, iindex, jindex, cellij ) 
          call Scell( k-1, 3, jindex, cell3j ) 
          rMulij = cellij + ((celli3-celli4/2)*cell3j) / d2
          b(i) = b(i) + rMulij * rates(j)
        enddo
      enddo
c     ..begin computing ak using Q11, the upper-left corner of sM2inv
c     ..begin computing ai, i=1,..,k-1, using rows 1, 4,... of Q11
      do i=1,k-1
        iindex = 3*i - 2 
        a(i) = 0
        call Scell( k-1, iindex, k3-3, celli3 ) 
        call Scell( k-1, iindex, k3-4, celli4 ) 
        rMuri2 = (q*celli3 - p*celli4  ) / d2
        rMuri1 = ( -celli3 +   celli4/2) / d2
        gammai = rMuri1 * rate0 + rMuri2 * rates(k)
        do j=1,k-1
          jindex= 3*j - 2
          call Scell( k-1, iindex, jindex, cellij ) 
          call Scell( k-1,      3, jindex, cell3j ) 
          rMulij = cellij + ((celli3 - celli4/2)*cell3j) / d2
          a(i) = a(i) + rMulij * rates(j)
          gammai = gammai + rMulij * rates(j)
        enddo
        top    = top    + rMuri2 * gammai
        bottom = bottom + rMuri2**2
      enddo
c     ..finish computing ak
      bottom = 3 + (bottom / 3)
      a(k)   = top / bottom
c     ..finish computing (ai, bi), i=1,...,k-1 using Q12
      do i=1,k-1
        do ioffset=0,1
          iindex = 3*i - 2 + ioffset
          call Scell( k-1, iindex, k3-3, celli3 )
          call Scell( k-1, iindex, k3-4, celli4 )
          rMuri2 = (q*celli3 - p*celli4  ) / d2
          rMuri1 = ( -celli3 +   celli4/2) / d2
          ab = rMuri2 * (rates(k) - a(k)/3) + rMuri1 * rate0
          if (ioffset .eq. 0) a(i) = a(i) + ab
          if (ioffset .eq. 1) b(i) = b(i) + ab
        enddo
      enddo
c     ..compute bk
      b(k) = 2*a(k-1) + b(k-1)
      return
      end


      subroutine pqfit3( k, rates, ratekp1, a, b )
c     Huifen Chen and Bruce Schmeiser, March 14, 2015.
c     PQRS Context: 3. Finite horizon, only right end-point rate.
c     Fit unconstrained rates; i.e., ignore negativity.
c     Method:  Time reverse Context 2.
c     input: 
c       k     = number of intervals
c       rates = specified rates, i=1,...,k
c       ratekp1 = right end-point rate
c       
c     output:
c       (a,b) = second- and first-order quadratic coefficients
      implicit doubleprecision (a-h,o-z)
      implicit integer (i-n)
      double precision, dimension (1:k) :: rates, a, b
      khalf = int( k/2 )
c     ..reverse time for input rates
      rate0 = ratekp1
      do i=1,khalf
        temp         = rates(i)
        rates(i)     = rates(k-i+1)                
        rates(k-i+1) = temp
      enddo  
c     ..call context 2 fitting
      call pqfit2( k, rates, rate0, a, b )
c     ..convert coefficients to x from 1-x
c       these coefficients are for interval k
      do i=1,k
        b(i) = -2*a(i) - b(i)
      enddo
c     ..reverse time for the input rates and results
      do i=1,khalf
        temp     = rates(i)
        rates(i) = rates(k-i+1)
        rates(k-i+1) = temp
        temp     = a(i)
        a(i)     = a(k-i+1)
        a(k-i+1) = temp
        temp     = b(i)
        b(i)     = b(k-i+1)
        b(k-i+1) = temp
      enddo
      return
      end


      subroutine pqfit4( k, rates, a, b )
c     Huifen Chen and Bruce Schmeiser, March 14, 2015.
c     PQRS Context: 4. Finite horizon, no end-point rates.
c     Fit unconstrained rates; i.e., ignore negativity.
c     Method:  Minimize sum_i=1^k a_i^2, subject to k means, 
c              k-1 continuity, and k-1 1st derivatives.
c              a(1) and a(k) are the free variables.
c     input: 
c       k     = number of intervals
c       rates = specified rates, i=1,...,k
c     output:
c       (a,b) = second- and first-order quadratic coefficients
      implicit doubleprecision (a-h,o-z)
      implicit integer (i-n)
      double precision, dimension (1:k) :: rates, a, b
      if (k .eq. 1) then
        a(1) = 0
        b(1) = 0
        return
      elseif (k .eq. 2) then
        a(1) = 0
        b(1) = rates(2) - rates(1)
        a(2) = 0
        b(2) = b(1)
        return
      endif
c     ..k > 2, so initialize
      k3    = 3*k 
      alpha = 1
      beta  = 0
      phi   = 0
      eta   = 1
      xi    = 0
c     ..compute R22, the 4x4 lower-right corner of rM4inv
      call Scell( k-2, 3, k3-6, w1 )
      call Scell( k-2, 3, k3-7, w2 )
      call Scell( k-2, 2, k3-6, w3 )
      call Scell( k-2, 2, k3-7, w4 )
      d4 = w1 - w2/2 - w3/2 + w4/4
      rMlr55 =  (-w2 + w4/2)            / d4 
      rMlr54 =   1                      / d4
      rMlr53 =  -1                      / d4
      rMlr52 =     0.5d0                / d4
      rMlr45 =  (w1 - w3/2)             / d4
      rMlr44 =    -0.5d0                / d4
      rMlr43 =     0.5d0                / d4
      rMlr42 =    -0.25d0               / d4
      rMlr35 =  (w1*w4 - w2*w3)         / d4
      rMlr34 =  (w3 - w4/2)             / d4
      rMlr33 =  (-w3 + w4/2)            / d4
      rMlr32 =  (w1 - w2/2)             / d4
      rMlr25 =  ((w2*w3 - w1*w4)/2)     / d4
      rMlr24 =  (w1 - w2/2 - w3 + w4/2) / d4
      rMlr23 =  (w3/2 - w4/4)           / d4
      rMlr22 =  (w2/4 - w1/2)           / d4
c     ..compute a1 and ak
      do i=1,k-2
        iindex = 3*i - 2
c       ..using R12, the upper-right-corner of rM4inv
        call Scell( k-2, iindex, k3-6, celli6 )
        call Scell( k-2, iindex, k3-7, celli7 )
        rMuri5 = rMlr55*celli6 + rMlr45*celli7
        rMuri4 = rMlr54*celli6 + rMlr44*celli7
        rMuri3 = rMlr53*celli6 + rMlr43*celli7
        rMuri2 = rMlr52*celli6 + rMlr42*celli7
        p = -rMuri4 / 3 - rMuri3 - 2*rMuri2
        q = -rMuri5 / 3
        r =  rMuri4 * rates(1) + rMuri5*rates(k)
        do j=2,k-1
c       ..using R11, the upper-left-corner of rM4inv
          jindex = 3*j - 5
          call Scell( k-2, iindex, jindex, cellij )
          call Scell( k-2, 2,      jindex, cell2j )
          call Scell( k-2, 3,      jindex, cell3j )
          rMulij = cellij + ((2*celli6-celli7)*(cell2j-2*cell3j))/(4*d4)
          r = r + rMulij * rates(j)
        enddo
        alpha = alpha + p**2
        beta  = beta  + p*q
        phi   = phi   + p*r
        eta   = eta   + q**2
        xi    = xi    + q*r
      enddo
      denom = beta**2 - alpha*eta
      a(1)  = (eta*phi  - beta*xi ) / denom
      a(k)  = (alpha*xi - beta*phi) / denom
      factor2 =          - 2*a(1)
      factor3 =          -   a(1)
      factor4 = rates(1) -   a(1)/3
      factor5 = rates(k) -   a(k)/3
c     ..compute b1 and bk using rows k3-3 and k3-5 of M4inv
c     ....from R22, the lower-right corner of M4inv
      b(1) =        rMlr32 * factor2 + rMlr33 * factor3
      b(1) = b(1) + rMlr34 * factor4 + rMlr35 * factor5
      b(k) =        rMlr52 * factor2 + rMlr53 * factor3
      b(k) = b(k) + rMlr54 * factor4 + rMlr55 * factor5
c     ....from R21, the lower-left corner of M4inv
      do j=2,k-1
        jindex = 3*j - 5
        call Scell( k-2, 2, jindex, cell2j )
        call Scell( k-2, 3, jindex, cell3j )
        rMll3j = rMlr33 * cell3j + rMlr32 * cell2j
        rMll5j = rMlr53 * cell3j + rMlr52 * cell2j
        b(1) = b(1) + rMll3j * rates(j)
        b(k) = b(k) + rMll5j * rates(j)
      enddo
c     ..compute (ai,bi), i=2,...,k-1
      do i=2,k-1
        do ioffset=0,1
          iindex = i*3 - 5 + ioffset
          call Scell( k-2, iindex, k3-6, celli6 )
          call Scell( k-2, iindex, k3-7, celli7 )
c         ....use R12, the upper-right corner of M4inv
          rMuri5 = celli6*rMlr55 + celli7*rMlr45
          rMuri4 = celli6*rMlr54 + celli7*rMlr44
          rMuri3 = celli6*rMlr53 + celli7*rMlr43
          rMuri2 = celli6*rMlr52 + celli7*rMlr42
          ab=rMuri5*factor5+rMuri4*factor4+rMuri3*factor3+rMuri2*factor2
          if (ioffset .eq. 0) a(i) = ab
          if (ioffset .eq. 1) b(i) = ab
c         ....use R11, the upper-left corner of M4inv
          do j=2,k-1
            jindex = 3*j - 5
            call Scell( k-2, iindex, jindex, cellij )
            call Scell( k-2, 2,      jindex, cell2j )
            call Scell( k-2, 3,      jindex, cell3j )
            rMulij = cellij+((2*celli6-celli7)*(cell2j-2*cell3j))/(4*d4)
            if (ioffset .eq. 0) a(i) = a(i) + rMulij * rates(j)
            if (ioffset .eq. 1) b(i) = b(i) + rMulij * rates(j)
          enddo
        enddo
      enddo
      return
      end


      subroutine MNOfit( icontxt, k, rates, rate0, ratekp1, 
     &                               a, b, c, aplus, bplus, cplus)
c     Fen Chen and Bruce Schmeiser, Feb 14, 2015.
c     For each interval, modify the quadratic ax^2+bx+c to
c     max{0, aplus x^2 + bplus x + cplus}, so then no negative rate.
c     input:
c       icontxt = context \in {0,1,2,3,4}
c       k = number of intervals
c       rates = specified rates, i=1,...,k
c       rate0, ratekp1 = time-zero and time-k specified rates
c       (a,b,c) = PQRS quadratic coefficients
c     output:
c       (aplus, bplus, cplus) = updated quadratic coefficients
      implicit doubleprecision (a-h,o-z)
      implicit integer (i-n)
      parameter (epsilon = 1d-10)
      double precision, dimension (1:k):: rates,a,b,c,aplus,bplus,cplus
      integer, dimension (1:k):: icase
c     ..set default values
      do i=1,k
        aplus(i) = a(i)
        bplus(i) = b(i)
        cplus(i) = c(i)
      enddo
c     ..determine the initial negativity case for each interval
      call pqcase( k, rates, a, b, c, icase )
c     ..MNO update in specific case order
      do i=1,k
        call MNOfit4( icontxt, k, rates, rate0, ratekp1, icase, i,
     &                               a, b, c, aplus, bplus, cplus)
      enddo
      do i=1,k
        call MNOfit3( icontxt, k, rates, rate0, ratekp1, icase, i,
     &                               a, b, c, aplus, bplus, cplus)
      enddo
      do i=1,k
        call MNOfit2( icontxt, k, rates, rate0, ratekp1, icase, i,
     &                               a, b, c, aplus, bplus, cplus)
      enddo
      do i=1,k
        call MNOfit1( icontxt, k, rates, rate0, ratekp1, icase, i,
     &                               a, b, c, aplus, bplus, cplus)
      enddo
        call MNOfit0( icontxt, k, rates, rate0, ratekp1, icase, 1,
     &                               a, b, c, aplus, bplus, cplus)
        call MNOfit0( icontxt, k, rates, rate0, ratekp1, icase, k,
     &                               a, b, c, aplus, bplus, cplus)
c     ..override with zero-rate values
      do i=1,k
        if (rates(i) .lt. epsilon) then
          aplus(i) = 0
          bplus(i) = 0
          cplus(i) = 0
        endif
      enddo
      return
      end


      subroutine MNOfit4( icontxt, k, rates, rate0, ratekp1, icase, i,
     &                                   a, b, c, aplus, bplus, cplus)
c     Fen Chen and Bruce Schmeiser, Feb 14, 2015
c     MNO update: max{ 0, aplus x^2 + bplus x + cplus }
c       Case 4---negative rates; 2 nonnegative end-point rates
c     input:
c       icontxt = context \in {0,1,2,3,4}
c       k = number of intervals
c       rates = specified rates, i=1,...,k
c       rate0, ratekp1 = time-zero and time-k specified rates
c       icase = negativity case {0,1,2,3,4}
c       (a,b,c) = PQRS quadratic coefficients
c     output:
c       (aplus, bplus, cplus) = MNO quadratic coefficients
      implicit doubleprecision (a-h,o-z)
      implicit integer (i-n)
      parameter (epsilon = 1d-10)
      double precision, dimension (1:k):: rates,a,b,c,aplus,bplus,cplus
      integer, dimension (1:k):: icase
      if (icase(i) .eq. 4) then
        qrate = rates(i)
        q0    = c(i)
        q1    = a(i) + b(i) + c(i)
        cplus(i) = c(i)
c       ..use secant method.
        aold = a(i)
        bold = q1 - cplus(i) - aold
        call qmaxint( aold, bold, q0, qintold )
        a2 = a(i) + 0.10*abs(a(i))       ! aplus >= a, always
        qintnew = qrate + 10*epsilon + 1 ! always 1st-time failure
        do while ( abs( qintnew - qrate ) .gt. epsilon )
          aplus(i) = a2
          bplus(i) = q1 - q0 - aplus(i)
          call qmaxint( aplus(i), bplus(i), q0, qintnew )
          if (qintnew .ne. qintold) then
            a2 = aplus(i) 
     &           - (aplus(i)-aold)*(qintnew-qrate)/(qintnew-qintold)
            aold = aplus(i)
            qintold = qintnew
          else
            qintnew = qrate
          endif
        enddo
      endif
      return
      end


      subroutine MNOfit2( icontxt, k, rates, rate0, ratekp1, icase, i,
     &                                a, b, c, aplus, bplus, cplus)
c      Fen Chen and Bruce Schmeiser, Feb 14, 2015
c     MNO update: max{ 0, aplus x^2 + bplus x + cplus }
c       Case 2---negative rates; only right nonnegative end-point rate
c     input:
c       icontxt = context \in {0,1,2,3,4}
c       k = number of intervals
c       rates = specified rates, i=1,...,k
c       rate0, ratekp1 = time-zero and time-k specified rates
c       icase = negativity case {0,1,2,3,4}
c       (a,b,c) = PQRS quadratic coefficients
c     output:
c       (aplus, bplus, cplus) = MNO quadratic coefficients
      implicit doubleprecision (a-h,o-z)
      implicit integer (i-n)
      double precision, dimension (1:k):: rates,a,b,c,aplus,bplus,cplus
      integer, dimension (1:k):: icase
      if (icase(i) .eq. 2) then
        qrate = rates(i)
        q1 = a(i) + b(i) + c(i)
c       ..if there is a next interval, use it for q1p
        if (icontxt .gt. 0 .and. i .eq. k) then
          q1p = 2*a(i) + b(i)
        else
          if (i .lt. k) then
            inext = i + 1
          elseif (icontxt .eq. 0) then
            inext = 1
          endif
          q1p = bplus(inext)
        endif
c       ..case 2 is closed form. 
        term1 = -q1 * (4*q1**2 - 9*q1p*qrate)
        term2 = 2*(2*q1**2 - 3*q1p*qrate)**3
        aplus(i)  = (term1 - sqrt( term2 ) ) / (18*qrate**2)          
        bplus(i)  = q1p - 2*aplus(i)
        cplus(i)  = q1 - q1p + aplus(i)
      endif
      return
      end


      subroutine MNOfit3( icontxt, k, rates, rate0, ratekp1, icase, i,
     &                                   a, b, c, aplus, bplus, cplus)
c      Fen Chen and Bruce Schmeiser, Feb 14, 2015
c     MNO update: max{ 0, aplus x^2 + bplus x + cplus }
c       Case 3---negative rates; only left nonnegative end-point rate
c     input:
c       icontxt = context \in {0,1,2,3,4}
c       k = number of intervals
c       rates = specified rates, i=1,...,k
c       rate0, ratekp1 = time-zero and time-k specified rates
c       icase = negativity case {0,1,2,3,4}
c       (a,b,c) = PQRS quadratic coefficients
c     output:
c       (aplus, bplus, cplus) = MNO quadratic coefficients
      implicit doubleprecision (a-h,o-z)
      implicit integer (i-n)
      double precision, dimension (1:k):: rates,a,b,c,aplus,bplus,cplus
      integer, dimension (1:k):: icase
      if (icase(i) .eq. 3) then
        qrate = rates(i)
        q0 = c(i)
c       ..if there is a previous interval, use it for q0p
        if (icontxt .gt. 0 .and. i .eq. 1) then
          q0p = b(i)
        else
          if (i .gt. 1) then
            iprev = i - 1
          elseif (icontxt .eq. 0) then
            iprev = k
          endif
          q0p = 2*aplus(iprev) + bplus(iprev)
        endif
c       ..case 3 is closed form. 
        term1 = -q0 * (4*q0**2 + 9*q0p*qrate)
        term2 = 2*(2*q0**2 + 3*q0p*qrate)**3
        aplus(i) = (term1 - sqrt( term2 ) ) / (18*qrate**2)          
        bplus(i) = q0p
        cplus(i) = c(i)
      endif
      return
      end


      subroutine MNOfit1( icontxt, k, rates, rate0, ratekp1, icase, i,
     &                                   a, b, c, aplus, bplus, cplus)
c      Fen Chen and Bruce Schmeiser, Feb 14, 2015
c     MNO update: max{ 0, aplus x^2 + bplus x + cplus }
c       Case 1---negative rates; no nonnegative end-point rate
c     input:
c       icontxt = context \in {0,1,2,3,4}
c       k = number of intervals
c       rates = specified rates, i=1,...,k
c       rate0, ratekp1 = time-zero and time-k specified rates
c       icase = negativity case {0,1,2,3,4}
c       (a,b,c) = PQRS quadratic coefficients
c     output:
c       (aplus, bplus, cplus) = MNO quadratic coefficients
      implicit doubleprecision (a-h,o-z)
      implicit integer (i-n)
      parameter (epsilon = 1d-10)
      double precision, dimension (1:k):: rates,a,b,c,aplus,bplus,cplus
      integer, dimension (1:k):: icase
      if (icase(i) .eq. 1) then
        qrate = rates(i)        ! same as for case 4
        q0 = c(i)
        q1 = a(i) + b(i) + c(i)
        cplus(i) = c(i)
c       ..use secant method.
        aold = a(i)
        bold = q1 - cplus(i) - aold
        call qmaxint( aold, bold, q0, qintold )
        a2 = a(i) + 0.10*abs(a(i))            ! aplus > a, always
          !for the future:a < aplus < (q0 + q1) - 2sqrt(q0q1) < 0
        qintnew = qrate + 10*epsilon + 1 ! always 1st-time failure
        do while ( abs( qintnew - qrate ) .gt. epsilon )
          aplus(i) = a2
          bplus(i) = q1 - q0 - aplus(i)
          call qmaxint( aplus(i), bplus(i), q0, qintnew )
          if (qintnew .ne. qintold) then
            a2 = aplus(i) 
     &           - (aplus(i)-aold)*(qintnew-qrate)/(qintnew-qintold)
            aold = aplus(i)
            qintold = qintnew
          else
            qintnew = qrate
          endif
        enddo
      endif
      return
      end


      subroutine MNOfit0( icontxt, k, rates, rate0, ratekp1, icase, i,
     &                                   a, b, c, aplus, bplus, cplus)
c     Fen Chen and Bruce Schmeiser, Feb 14, 2015
c     MNO update: max{ 0, aplus x^2 + bplus x + cplus }
c       Case 0---no negative rates; finite horizon, context 2, 3, 4
c         interval {1,k} adjacent to an updated case-4 interval
c     input:
c       icontxt = context \in {0,1,2,3,4}
c       k = number of intervals
c       rates = specified rates, i=1,...,k
c       rate0, ratekp1 = time-zero and time-k specified rates
c       icase = negativity case {0,1,2,3,4}
c       (a,b,c) = PQRS quadratic coefficients
c     output:
c       (aplus, bplus, cplus) = MNO quadratic coefficients
      implicit doubleprecision (a-h,o-z)
      implicit integer (i-n)
      double precision, dimension (1:k):: rates,a,b,c,aplus,bplus,cplus
      integer, dimension (1:k):: icase
      if (i .eq. 1) then
        if (icontxt .eq. 4 .or. icontxt .eq. 3) then
c         ..first interval        
          if (icase(1) .eq. 0 .and. icase(2) .eq. 4) then
            qrate = rates(1)
            q1  = cplus(2)
            q1p = bplus(2)
            cplus(1) = 3*qrate -2*q1 + q1p/2
            bplus(1) = 2*(q1 - cplus(1)) - q1p
            aplus(1) = (q1p - bplus(1)) / 2
            call pqcase( k, rates, aplus, bplus, cplus, icase )
c           ..re-update if interval 1's case changed
            i = 1
            if (icase(1) .eq. 2) then
              call MNOfit2( icontxt, k, rates, rate0, ratekp1, icase, i,
     &                                     a, b, c, aplus, bplus, cplus)
            elseif (icase(1) .eq. 4) then
              call MNOfit4( icontxt, k, rates, rate0, ratekp1, icase, i,
     &                                     a, b, c, aplus, bplus, cplus)
            endif
          endif
        endif
      elseif (i .eq. k) then
        if (icontxt .eq. 4 .or. icontxt .eq. 2) then
c         ..last interval
          if (icase(k) .eq. 0 .and. icase(k-1) .eq. 4) then
            qrate = rates(k)
            q0    = cplus(k)
            q0p   = 2*aplus(k-1) + bplus(k-1)
            cplus(k) = q0
            bplus(k) = q0p
            aplus(k) = 3*(qrate - bplus(k)/2 - cplus(k))
            call pqcase( k, rates, aplus, bplus, cplus, icase )
c           ..re-update if interval k's case changed
            if (icase(k) .eq. 3) then
              call MNOfit3( icontxt, k, rates, rate0, ratekp1, icase, k,
     &                                     a, b, c, aplus, bplus, cplus)
              elseif (icase(k) .eq. 4) then
              call MNOfit4( icontxt, k, rates, rate0, ratekp1, icase, k,
     &                                     a, b, c, aplus, bplus, cplus)
            endif
          endif
        endif
      endif
      return
      end


      subroutine pqcase( k, rates, a, b, c, icase )
c     Fen Chen and Bruce Schmeiser, Feb 14, 2015.
c     For each interval, determine the negativity case.
c     input:
c       k = number of intervals
c       rates = specified rates, i=1,...,k
c       (a,b,c) = PQRS quadratic coefficients
c     output:
c       icase = the negativity case for i=1,...,k
      implicit doubleprecision (a-h,o-z)
      implicit integer (i-n)
      double precision, dimension (1:k):: rates, a, b, c
      integer, dimension (1:k):: icase
      do i=1,k
c       ..compute qmin = min{ ax^2 + bx + c : x \in [0,1] }
        q0 = c(i)
        q1 = a(i) + b(i) + c(i)
        qmin = min( q0, q1 )
        if ( a(i) .gt. 0 ) then
          xstar = -b(i) / (2*a(i))
          if (xstar .gt. 0 .and. xstar .lt. 1) then
            qmin = c(i) - (b(i)**2 / (4*a(i)))
          endif
        endif
c       ..compute negativity cases
        icase(i) = 0
        if (qmin .lt. 0) then
          if (q0 .ge. 0 .and. q1 .ge. 0) then
            icase(i) = 4
          elseif (q0 .ge. 0 .and. q1 .lt. 0) then
            icase(i) = 3
          elseif (q0 .lt. 0 .and. q1 .ge. 0) then
            icase(i) = 2
          else
            icase(i) = 1        
          endif
        endif
      enddo
      return
      end


      subroutine qmaxint( a, b, c, qrate )
c     Huifen Chen and Bruce Schmeiser. Feb 7, 2015
c     purpose: integrate max{ 0, a x^2 + b x + c } over [0,1].
c     Known to be positive only on (alpha, beta) and (gamma, 1).
      implicit doubleprecision (a-h,o-z)
      implicit integer (i-n)
c     input:
c       a, b, c = quadratic coeffients
c     output:
c       qrate = integral value
      call qpartit( a, b, c, alpha, beta, gamma )
      qrate = (a/3) * ( (beta**3 - alpha**3) + (1 - gamma**3) )
     &        + (b/2) * ( (beta**2 - alpha**2) + (1 - gamma**2) )
     &        +  c    * ( (beta    - alpha   ) + (1 - gamma   ) )
      return
      end


      subroutine qpartit( a, b, c, alpha, beta, gamma )
c     Huifen Chen and Bruce Schmeiser.  March 1, 2014
c     purpose: Over [0,1], partition q(x) = ax^2 + bx + c 
c       into negative and non-negative intervals.
c     input: a, b, c
c     output: 
c       (alpha, beta, gamma) = the sign-change points in [0,1].   
c         Always    q(0,    alpha) <  0.
c                   q[alpha, beta] >= 0.
c                   q(beta, gamma) <  0.
c                   q[gamma,    1] >= 0.
c         The three values are not unique; we use
c         alpha = 0 when a > 0 and gamma = 1 when a < 0.
      implicit doubleprecision (a-h,o-z)
      implicit integer (i-n)
      alpha = 0
      beta  = 0
      gamma = 1
      if (a .eq. 0) then
        if (b .eq. 0) then         
          if (c .ge. 0) beta = 1
        else
          root = - c / b   
          if (b .lt. 0) then
            beta  = max( 0d0, min( 1d0, root ) )
          else
            gamma = max( 0d0, min( 1d0, root ) )
          endif
        endif
      else
        char = b**2 - 4*a*c
        if (char .lt. 0) then
          if (a .ge. 0) beta = 1
        else
          schar = sqrt( char )   
          root1 = (- b - schar) / (2*a)      
          root2 = (- b + schar) / (2*a)      
          beta  = max( 0d0, min( 1d0, root1 ) )
          if (a .lt. 0) then
            alpha = max( 0d0, min( 1d0, root2 ) )
          else           
            gamma = max( 0d0, min( 1d0, root2 ) )
          endif
        endif
      endif
      return
      end         
